You can search for identifiers within the package.
in-package search v0.2.0
bimage
Bimage.Input
Defines the type used as input to operations
type ('a, 'b, 'c) t = ('a, 'b, 'c) Image.t array
val get : ('a, 'b, 'c) t -> int -> ('a, 'b, 'c) Image.t
Get an image from the input, raising Error.Exc (`Invalid_input index) if the provided index is out of bounds.
Error.Exc (`Invalid_input index)
val make_output : ?width:int -> ?height:int -> ('a, 'b, 'c) t -> ('a, 'b, 'c) Image.t
Create an output image width the given width and height if provided, otherwise the generated image will match the first input image in size, kind and color